!**************************************************************************************************
! Module                      | metabufr_module
! Author                      | F. Guillaume (CNRM/GMAP/OBS)
! Creation Date               | 25/04/2016
!--------------------------------------------------------------------------------------------------
! Description : subroutines for counting & splitting bufr files
!--------------------------------------------------------------------------------------------------
! Global Variables : 
!--------------------------------------------------------------------------------------------------
! History :
!**************************************************************************************************

MODULE Metabufr_module

use            :: parkind1,     only : jpim, jpib, jprb
use            :: bufr_module

implicit none
private

public :: CountBufr, SplitFile, SplitSubsets, Erreur

 integer(kind=jpim)  :: iword
 integer(kind=jpim)  :: kel
 character(len=1024) :: ErrString

contains

!**************************************************************************************************
! Subroutine/Function/Program | CountBufr
! Author                      | F. Guillaume (CNRM/GMAP/OBS)
! Creation Date               | 30/05/2016
!--------------------------------------------------------------------------------------------------
! Description : count number of single bufr files & subsets contained in a 'meta'bufr file
!--------------------------------------------------------------------------------------------------
! Global Variables : 
!--------------------------------------------------------------------------------------------------
! History :
!**************************************************************************************************
SUBROUTINE CountBufr(InputFile,NbFic,NbSubset)
 character(len=128),INTENT(IN)  :: InputFile
 integer(kind=jpim),INTENT(OUT) :: NbFic
 integer(kind=jpim),INTENT(OUT),OPTIONAL :: NbSubset
 integer(kind=jpim)             :: nulbufr, iret, iret2

nulbufr  = 0
iret     = 0
iret2    = 0
nbfic    = 0
if (present(nbsubset)) NbSubset = 0

call pbopen(nulbufr,InputFile,'r',iret)
if (iret /= 0 ) call erreur("***** ERROR : can't open input file.")
do while (iret == 0)
  call pbbufr(nulbufr,kbuff,jbyte,kbufl,iret)
  if (iret /= 0) cycle
  NbFic = NbFic + 1
  iword = kbufl / (jbpw / 8) + 1
  call bus012(iword,kbuff,ksup,ksec0,ksec1,ksec2,iret2)
  if (iret2 /= 0) then
    write(ErrString,'("***** ERROR bus012() : Bufr file n° ",I6.6," is incorrect.")') NbFic
    print *, ErrString
!   call erreur(trim(ErrString))
  else
    if (present(NbSubset)) NbSubset = NbSubset + ksup(6)
  endif
enddo
if (iret /= -1) stop 3
call pbclose(nulbufr,iret)
END SUBROUTINE CountBufr

!**************************************************************************************************
! Subroutine/Function/Program | SpliFile
! Author                      | F. Guillaume (CNRM/GMAP/OBS)
! Creation Date               | 30/05/2016
!--------------------------------------------------------------------------------------------------
! Description : Split 'meta'bufr file to single multi-subset bufr files
!--------------------------------------------------------------------------------------------------
! Global Variables : 
!--------------------------------------------------------------------------------------------------
! History :
!**************************************************************************************************
SUBROUTINE SplitFile(InputFile,OutputFile,FirstFile,LastFile)
 character(len=128),INTENT(IN)  :: InputFile, OutputFile
 integer(kind=jpim),INTENT(IN)  :: FirstFile
 integer(kind=jpim),INTENT(INOUT)  :: LastFile
 character(len=140)             :: OutputFileName
 character(len=6)    :: temporaire
 integer(kind=jpim)             :: ret, NbFic, CurrentFile

ret      = 0
values   = rvind

call CountBufr(InputFile,NbFic)
if (NbFic > 999999) call erreur("***** ERROR : too many single bufr files contained in input file (>999999).")
if (LastFile > NbFic) call erreur("***** ERROR : number of bufr files exceed NbFic.")

if (FirstFile == 0) then
  CurrentFile = 1
  LastFile    = NbFic
else
  CurrentFile = FirstFile
endif

do while (ret == 0 .and. CurrentFile <= LastFile)
  call ReadBufr(InputFile,CurrentFile,ret)
  if (ret /=0) cycle
  write(temporaire,'(I6.6)') Currentfile
  OutputFileName = trim(adjustl(OutputFile))//trim(adjustl(temporaire))
 ! write(OutputFileName,'(A,"_",i6)') trim(adjustl(OutputFile)), CurrentFile
  call WriteBufr(OutputFileName,ret)
  if (ret == 0) CurrentFile = CurrentFile + 1
enddo
END SUBROUTINE SplitFile

!**************************************************************************************************
! Subroutine/Function/Program | SplitSubsets
! Author                      | F. Guillaume (CNRM/GMAP/OBS)
! Creation Date               | 30/05/2016
!--------------------------------------------------------------------------------------------------
! Description : Split 'meta'bufr file to single (1 subset) bufr files
!--------------------------------------------------------------------------------------------------
! Global Variables : 
!--------------------------------------------------------------------------------------------------
! History : !RESTE LE CHOIX DES SUBSETS A ECLATER... LE RESTE FONCTIONNE
!**************************************************************************************************
SUBROUTINE SplitSubsets(InputFile,OutputFile,FirstFile,LastFile,FirstSubset,LastSubset)

 character(len=128),INTENT(IN)    :: InputFile, OutputFile
 integer(kind=jpim),INTENT(IN)    :: FirstFile, FirstSubset, LastSubset
 integer(kind=jpim),INTENT(INOUT) :: LastFile
 character(len=140)               :: OutputFileName
 integer(kind=jpim)               :: ret, NbFic, NbSubset, CurrentFile
 integer(kind=jpim)               :: i

values   = rvind
ret      = 0

call CountBufr(InputFile,NbFic,NbSubset)
if ((NbSubset/NbFic) > 9999) call erreur("***** ERROR : too many subsets per file (>9999).")
if (LastFile > NbFic) call erreur("***** ERROR : number of bufr files exceed NbFic.")

if (FirstFile == 0) then
  CurrentFile = 1
  LastFile    = NbFic
else
  CurrentFile = FirstFile
endif

print *,"" 

do while (ret == 0 .and. CurrentFile <= LastFile)
  call ReadBufr(InputFile,CurrentFile,ret)
  if (ret /=0) cycle
  NbSubset = ksup(6)
  ksec3(3) = 1
  ksup(6)  = 1

  if (FirstSubset == 0) then
    do i=1, NbSubset
      values(1:kel) = values((i - 1) * kel + 1:i * kel)
      write(OutputFileName,'(A,"_",i6.6,"_",i4.4)') trim(adjustl(OutputFile)), CurrentFile, i
      call WriteBufr(OutputFileName,ret)
      if (ret /= 0) call erreur("***** ERROR : I can't write bufr file")
    enddo
  else
    if (FirstSubset > NbSubset) then
      write(*,'("***** WARNING : First subset asked does not exist in bufr file ",I6," which is skipped")') CurrentFile
      CurrentFile = CurrentFile + 1
      cycle
    else
      if (LastSubset > NbSubset) write(*,'("***** WARNING : bufr ",I6," contains only ",I4," subsets.")') CurrentFile, NbSubset
      do i=FirstSubset, min(NbSubset,LastSubset)
        values(1:kel) = values((i - 1) * kel + 1:i * kel)
        write(OutputFileName,'(A,"_",i6.6,"_",i4.4)') trim(adjustl(OutputFile)), CurrentFile, i
        call WriteBufr(OutputFileName,ret)
        if (ret /= 0) call erreur("***** ERROR : I can't write bufr file")
      enddo
    endif
  endif
  if (ret == 0) CurrentFile = CurrentFile + 1
enddo
END SUBROUTINE SplitSubsets


!**************************************************************************************************
! Subroutine/Function/Program | ReadBufr
! Author                      | F. Guillaume (CNRM/GMAP/OBS)
! Creation Date               | 30/05/2016
!--------------------------------------------------------------------------------------------------
! Description : read a bufr file
!--------------------------------------------------------------------------------------------------
! Global Variables : 
!--------------------------------------------------------------------------------------------------
! History :
!**************************************************************************************************
SUBROUTINE ReadBufr(InputFile,NbFic,ret)
 character(len=128),INTENT(IN)    :: InputFile
 integer(kind=jpim),INTENT(IN)    :: nbfic
 integer(kind=jpim),INTENT(OUT)   :: ret
 integer(kind=jpim)               :: nulbufr, iret
 integer(kind=jpim)               :: i, j

nulbufr  = 0
iret     = 0
iword    = 0
kel      = 0
values   = rvind
kdata    = 0
ret      = 0

call pbopen(nulbufr,InputFile,'r',iret)
if (iret == 0 ) then
  do i=1, NbFic
    CALL pbbufr(nulbufr,kbuff,jbyte,kbufl,iret)
  enddo
  if (iret == 0) then
    iword = kbufl / (jbpw / 8) + 1
    call bus012(iword,kbuff,ksup,ksec0,ksec1,ksec2,iret)
    if (iret == 0 ) then
      kel = jwork / ksup(6)
      call bufrex(kbufl,kbuff,ksup,ksec0,ksec1,ksec2,ksec3,ksec4, &
               & kel,cnames,cunits,jwork,values,cvals,iret)
      if (iret == 0 ) then
        call busel(ktdlen,ktdlst,ktdexl,ktdexp,iret)
        if (iret == 0 ) then
          j = 1
          do i=1, ktdexl
            select case (ktdexp(i))
            case (203000:203255)
              call erreur("SORRY : I don't know how to handle 203YYY descriptors !")
            case (31000:31255)
              kdata(j) = nint(values(i))
              j        = j + 1
            end select
          enddo
        else
          ret = 5
        endif
      else
        ret = 4
      endif
    else
      ret = 3
    endif
  else
    ret = 2
  endif
  call pbclose(nulbufr,iret)
else
  ret = 1
endif
END SUBROUTINE ReadBufr


!**************************************************************************************************
! Subroutine/Function/Program | WriteBufr
! Author                      | F. Guillaume (CNRM/GMAP/OBS)
! Creation Date               | 30/05/2016
!--------------------------------------------------------------------------------------------------
! Description : Write a bufr file
!--------------------------------------------------------------------------------------------------
! Global Variables : 
!--------------------------------------------------------------------------------------------------
! History :
!**************************************************************************************************
SUBROUTINE WriteBufr(OutputFileName,ret)
 character(len=140),INTENT(IN) :: OutputFileName
 integer(kind=jpim),INTENT(OUT):: ret
 integer(kind=jpim)            :: nulbufr, iret
 integer(kind=jpim)            :: ilen
 
ret = 0 
 
call pbopen(nulbufr,OutputFileName,'w',iret)
if (iret == 0 ) then
  call bufren( ksec0,ksec1,ksec2,ksec3,ksec4,  &
             & ktdlen,ktdlst,kelem,kdata,kel,  &
             & jwork,values,cvals,kbufl,kbuff,iret)
  if (iret == 0 ) then 
    ilen = kbufl * jbpw / 8
    call pbwrite(nulbufr,kbuff,ilen,iret)
    if ( iret /= ilen) then
      write(*,'(I9," bytes writen <> ",I9," bytes wished.")') iret, ilen
      ret = 1
    endif
  else
    ret = 1
  endif
else
  ret = 1
endif
call pbclose(nulbufr,iret)
END SUBROUTINE WriteBufr


!**************************************************************************************************
! Subroutine/Function/Program | Erreur
! Author                      | F. Guillaume (CNRM/GMAP/OBS)
! Creation Date               | 30/05/2016
!--------------------------------------------------------------------------------------------------
! Description : Manage error codes
!--------------------------------------------------------------------------------------------------
! Global Variables : 
!--------------------------------------------------------------------------------------------------
! History :
!**************************************************************************************************
subroutine Erreur(message)
  character(len=*), intent(in) :: message

  write(*,'(/,A,/)') message
  print *, "Syntax : SplitBufr -i [path]InputFile -o [path)OutputFilePrefix [-f n[-m]] [-s n[-m]]"
  print *, "      or SplitBufr -i [path]InputFile -c"
  print *, "-f is used to specify a particular or a sequence of bufr files (n > 0 & m >= n)"
  print *, "-s is used to specify a particular or a sequence of subsets (n > 0 & m >= n)"
  print *, "-c is used to only count number of files and subsets contained in InputFile."
  stop
end subroutine erreur

END MODULE Metabufr_module

